@@ -52,25 +52,24 @@
 #include <osmocom/gsm/gsm_utils.h>
 
 #include "debug.h"
-#include <osmocom/sgsn/gb_proxy.h>
+#include <osmocom/gbproxy/gb_proxy.h>
 
 #include <osmocom/gsm/protocol/gsm_04_08_gprs.h>
 
 extern void *tall_sgsn_ctx;
 
 static const struct rate_ctr_desc global_ctr_description[] = {
-	{ "inv-bvci",	    "Invalid BVC Identifier          " },
-	{ "inv-lai",	    "Invalid Location Area Identifier" },
-	{ "inv-rai",	    "Invalid Routing Area Identifier " },
-	{ "inv-nsei",	    "No BVC established for NSEI     " },
-	{ "proto-err:bss",  "BSSGP protocol error      (BSS )" },
-	{ "proto-err:sgsn", "BSSGP protocol error      (SGSN)" },
-	{ "not-supp:bss",   "Feature not supported     (BSS )" },
-	{ "not-supp:sgsn",  "Feature not supported     (SGSN)" },
-	{ "restart:sgsn",   "Restarted RESET procedure (SGSN)" },
-	{ "tx-err:sgsn",    "NS Transmission error     (SGSN)" },
-	{ "error",          "Other error                     " },
-	{ "mod-peer-err",   "Patch error: no peer            " },
+	[GBPROX_GLOB_CTR_INV_BVCI]		= { "inv-bvci",	    "Invalid BVC Identifier          " },
+	[GBPROX_GLOB_CTR_INV_LAI]		= { "inv-lai",	    "Invalid Location Area Identifier" },
+	[GBPROX_GLOB_CTR_INV_RAI]		= { "inv-rai",	    "Invalid Routing Area Identifier " },
+	[GBPROX_GLOB_CTR_INV_NSEI]		= { "inv-nsei",	    "No BVC established for NSEI     " },
+	[GBPROX_GLOB_CTR_PROTO_ERR_BSS]		= { "proto-err:bss",  "BSSGP protocol error      (BSS )" },
+	[GBPROX_GLOB_CTR_PROTO_ERR_SGSN]	= { "proto-err:sgsn", "BSSGP protocol error      (SGSN)" },
+	[GBPROX_GLOB_CTR_NOT_SUPPORTED_BSS]	= { "not-supp:bss",   "Feature not supported     (BSS )" },
+	[GBPROX_GLOB_CTR_NOT_SUPPORTED_SGSN]	= { "not-supp:sgsn",  "Feature not supported     (SGSN)" },
+	[GBPROX_GLOB_CTR_RESTART_RESET_SGSN]	= { "restart:sgsn",   "Restarted RESET procedure (SGSN)" },
+	[GBPROX_GLOB_CTR_TX_ERR_SGSN]		= { "tx-err:sgsn",    "NS Transmission error     (SGSN)" },
+	[GBPROX_GLOB_CTR_OTHER_ERR]		= { "error",          "Other error                     " },
 };
 
 static const struct rate_ctr_group_desc global_ctrg_desc = {
@@ -145,7 +144,7 @@ static int gbprox_relay2sgsn(struct gbproxy_config *cfg, struct msgb *old_msg,
 		       PRIM_OP_REQUEST, msg);
 	rc = gprs_ns2_recv_prim(nsi, &nsp.oph);
 	if (rc < 0)
-		rate_ctr_inc(&cfg->ctrg->ctr[GBPROX_GLOB_CTR_TX_ERR_SGSN]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, GBPROX_GLOB_CTR_TX_ERR_SGSN));
 	return rc;
 }
 #endif
@@ -221,7 +220,7 @@ static int gbprox_relay2nse(struct msgb *old_msg, struct gbproxy_nse *nse,
 	rc = bssgp2_nsi_tx_ptp(nsi, nse->nsei, ns_bvci, msg, tlli);
 	/* FIXME: We need a counter group for gbproxy_nse */
 	//if (rc < 0)
-	//	rate_ctr_inc(&bvc->ctrg->ctr[GBPROX_PEER_CTR_TX_ERR]);
+	//	rate_ctr_inc(rate_ctr_group_get_ctr(bvc->ctrg, GBPROX_PEER_CTR_TX_ERR));
 
 	return rc;
 }
@@ -236,7 +235,7 @@ static int gbprox_relay2peer(struct msgb *old_msg, struct gbproxy_bvc *bvc,
 
 	rc = gbprox_relay2nse(old_msg, nse, ns_bvci);
 	if (rc < 0)
-		rate_ctr_inc(&bvc->ctrg->ctr[GBPROX_PEER_CTR_TX_ERR]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(bvc->ctrg, GBPROX_PEER_CTR_TX_ERR));
 
 	return rc;
 }
@@ -265,22 +264,24 @@ static struct gbproxy_sgsn *gbproxy_select_sgsn(struct gbproxy_config *cfg, cons
 	bool null_nri = false;
 
 	if (!tlli) {
-		sgsn = llist_first_entry(&cfg->sgsns, struct gbproxy_sgsn, list);
+		sgsn = gbproxy_sgsn_by_available(cfg);
 		if (!sgsn) {
+			LOGP(DGPRS, LOGL_ERROR, "Could not find any available SGSN\n");
 			return NULL;
 		}
-		LOGPSGSN(sgsn, LOGL_INFO, "Could not get TLLI, using first SGSN\n");
+		LOGPSGSN(sgsn, LOGL_INFO, "Could not get TLLI, using first available SGSN\n");
 		return sgsn;
 	}
 
 	if (cfg->pool.nri_bitlen == 0) {
 		/* Pooling is disabled */
-		sgsn = llist_first_entry(&cfg->sgsns, struct gbproxy_sgsn, list);
+		sgsn = gbproxy_sgsn_by_available(cfg);
 		if (!sgsn) {
+			LOGP(DGPRS, LOGL_ERROR, "Could not find any available SGSN\n");
 			return NULL;
 		}
 
-		LOGPSGSN(sgsn, LOGL_INFO, "Pooling disabled, using first configured SGSN\n");
+		LOGPSGSN(sgsn, LOGL_INFO, "Pooling disabled, using first available SGSN\n");
 	} else {
 		/* Pooling is enabled, try to use the NRI for routing to an SGSN
 		 * See 3GPP TS 23.236 Ch. 5.3.2 */
@@ -379,6 +380,31 @@ static int gbprox_bss2sgsn_tlli(struct gbproxy_cell *cell, struct msgb *msg, con
 	return gbprox_relay2peer(msg, sgsn_bvc, sig_bvci ? 0 : sgsn_bvc->bvci);
 }
 
+static int gbproxy_decode_bssgp(const struct bssgp_normal_hdr *bgph, int msg_len, struct tlv_parsed *tp, const char *log_pfx)
+{
+	int rc;
+
+	/* UNITDATA PDUs have a different header than the other PDUs */
+	if (bgph->pdu_type == BSSGP_PDUT_UL_UNITDATA || bgph->pdu_type == BSSGP_PDUT_DL_UNITDATA) {
+		const struct bssgp_ud_hdr *budh = (struct bssgp_ud_hdr *) bgph;
+		if (msg_len < sizeof(*budh))
+			return -OSMO_TLVP_ERR_MAND_IE_MISSING;
+		rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, tp, 1, budh->pdu_type, budh->data,
+					 msg_len - sizeof(*budh), 0, 0, DGPRS, log_pfx);
+		/* populate TLLI from the fixed headser into the TLV-parsed array so later code
+		 * doesn't have to worry where the TLLI came from */
+		tp->lv[BSSGP_IE_TLLI].len = 4;
+		tp->lv[BSSGP_IE_TLLI].val = (const uint8_t *) &budh->tlli;
+	} else {
+		rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, tp, 1, bgph->pdu_type, bgph->data,
+					 msg_len - sizeof(*bgph), 0, 0, DGPRS, log_pfx);
+	}
+
+	return rc;
+}
+
+static int gbproxy_tlli_from_status_pdu(struct tlv_parsed *tp, uint32_t *tlli, char *log_pfx);
+
 /* Receive an incoming PTP message from a BSS-side NS-VC */
 static int gbprox_rx_ptp_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uint16_t ns_bvci)
 {
@@ -416,23 +442,9 @@ static int gbprox_rx_ptp_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 		return tx_status(nse, ns_bvci, BSSGP_CAUSE_UNKNOWN_BVCI, &ns_bvci, msg);
 	}
 
-	/* UL_UNITDATA has a different header than all other uplink PDUs */
-	if (bgph->pdu_type == BSSGP_PDUT_UL_UNITDATA) {
-		const struct bssgp_ud_hdr *budh = (struct bssgp_ud_hdr *) msgb_bssgph(msg);
-		if (msgb_bssgp_len(msg) < sizeof(*budh))
-			return tx_status(nse, ns_bvci, BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
-		rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, &tp, 1, bgph->pdu_type, budh->data,
-					 msgb_bssgp_len(msg) - sizeof(*budh), 0, 0, DGPRS, log_pfx);
-		/* populate TLLI from the fixed headser into the TLV-parsed array so later code
-		 * doesn't have to worry where the TLLI came from */
-		tp.lv[BSSGP_IE_TLLI].len = 4;
-		tp.lv[BSSGP_IE_TLLI].val = (const uint8_t *) &budh->tlli;
-	} else {
-		rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, &tp, 1, bgph->pdu_type, bgph->data,
-					 msgb_bssgp_len(msg) - sizeof(*bgph), 0, 0, DGPRS, log_pfx);
-	}
+	rc = gbproxy_decode_bssgp(bgph, msgb_bssgp_len(msg), &tp, log_pfx);
 	if (rc < 0) {
-		rate_ctr_inc(&nse->cfg->ctrg->ctr[GBPROX_GLOB_CTR_PROTO_ERR_BSS]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(nse->cfg->ctrg, GBPROX_GLOB_CTR_PROTO_ERR_BSS));
 		return tx_status_from_tlvp(nse, rc, msg);
 	}
 	/* hack to get both msg + tlv_parsed passed via osmo_fsm_inst_dispatch */
@@ -482,7 +494,7 @@ static int gbprox_rx_ptp_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 		const uint8_t *mi_data = TLVP_VAL(&tp, BSSGP_IE_IMSI);
 		uint8_t mi_len = TLVP_LEN(&tp, BSSGP_IE_IMSI);
 		osmo_mobile_identity_decode(&mi, mi_data, mi_len, false);
-		nse = gbproxy_nse_by_imsi(nse->cfg, mi.imsi);
+		nse = gbproxy_nse_by_imsi(nse->cfg, mi.imsi, CACHE_USAGE_PAGING);
 		if (nse) {
 			OSMO_ASSERT(nse->sgsn_facing);
 			rc = gbprox_relay2nse(msg, nse, ns_bvci);
@@ -495,12 +507,28 @@ static int gbprox_rx_ptp_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 		osmo_fsm_inst_dispatch(bss_bvc->fi, BSSGP_BVCFSM_E_RX_FC_BVC, msg);
 		break;
 	case BSSGP_PDUT_STATUS:
-		/* TODO: Implement by inspecting the contained PDU */
-		if (!TLVP_PRESENT(&tp, BSSGP_IE_PDU_IN_ERROR))
+	{
+		struct gbproxy_sgsn *sgsn;
+		/* Check if the status needs to be terminated locally */
+		uint8_t cause = *TLVP_VAL(&tp, BSSGP_IE_CAUSE);
+
+		LOGPNSE(nse, LOGL_NOTICE, "Rx STATUS cause=0x%02x(%s)\n", cause,
+			bssgp_cause_str(cause));
+
+		if (gbproxy_tlli_from_status_pdu(&tp, &tlli, log_pfx) == 0)
+			sgsn = gbproxy_select_sgsn(nse->cfg, &tlli);
+		else
+			sgsn = gbproxy_select_sgsn(nse->cfg, NULL);
+
+		if (!sgsn) {
+			rc = -EINVAL;
 			break;
-		LOGPBVC(bss_bvc, LOGL_ERROR, "Rx %s: Implementation missing\n", pdut_name);
+		}
+
+		rc = gbprox_relay2nse(msg, sgsn->nse, ns_bvci);
 		break;
 	}
+	}
 
 	return 0;
 }
@@ -538,33 +566,19 @@ static int gbprox_rx_ptp_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 	if (!sgsn_bvc) {
 		LOGP(DGPRS, LOGL_NOTICE, "%s %s - Didn't find BVC for for PTP message, discarding\n",
 		     log_pfx, pdut_name);
-		rate_ctr_inc(&nse->cfg->ctrg-> ctr[GBPROX_GLOB_CTR_INV_BVCI]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(nse->cfg->ctrg, GBPROX_GLOB_CTR_INV_BVCI));
 		return tx_status(nse, ns_bvci, BSSGP_CAUSE_UNKNOWN_BVCI, &ns_bvci, msg);
 	}
 
 	if (!bssgp_bvc_fsm_is_unblocked(sgsn_bvc->fi)) {
 		LOGPBVC(sgsn_bvc, LOGL_NOTICE, "Rx %s: Dropping on blocked BVC\n", pdut_name);
-		rate_ctr_inc(&sgsn_bvc->ctrg->ctr[GBPROX_PEER_CTR_DROPPED]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(sgsn_bvc->ctrg, GBPROX_PEER_CTR_DROPPED));
 		return tx_status(nse, ns_bvci, BSSGP_CAUSE_BVCI_BLOCKED, &ns_bvci, msg);
 	}
 
-	/* DL_UNITDATA has a different header than all other uplink PDUs */
-	if (bgph->pdu_type == BSSGP_PDUT_DL_UNITDATA) {
-		const struct bssgp_ud_hdr *budh = (struct bssgp_ud_hdr *) msgb_bssgph(msg);
-		if (msgb_bssgp_len(msg) < sizeof(*budh))
-			return tx_status(nse, ns_bvci, BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
-		rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, &tp, 1, bgph->pdu_type, budh->data,
-					 msgb_bssgp_len(msg) - sizeof(*budh), 0, 0, DGPRS, log_pfx);
-		/* populate TLLI from the fixed headser into the TLV-parsed array so later code
-		 * doesn't have to worry where the TLLI came from */
-		tp.lv[BSSGP_IE_TLLI].len = 4;
-		tp.lv[BSSGP_IE_TLLI].val = (const uint8_t *) &budh->tlli;
-	} else {
-		rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, &tp, 1, bgph->pdu_type, bgph->data,
-					 msgb_bssgp_len(msg) - sizeof(*bgph), 0, 0, DGPRS, log_pfx);
-	}
+	rc = gbproxy_decode_bssgp(bgph, msgb_bssgp_len(msg), &tp, log_pfx);
 	if (rc < 0) {
-		rate_ctr_inc(&nse->cfg->ctrg->ctr[GBPROX_GLOB_CTR_PROTO_ERR_BSS]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(nse->cfg->ctrg, GBPROX_GLOB_CTR_PROTO_ERR_SGSN));
 		return tx_status_from_tlvp(nse, rc, msg);
 	}
 	/* hack to get both msg + tlv_parsed passed via osmo_fsm_inst_dispatch */
@@ -584,7 +598,7 @@ static int gbprox_rx_ptp_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 		const uint8_t *mi_data = TLVP_VAL(&tp, BSSGP_IE_IMSI);
 		uint8_t mi_len = TLVP_LEN(&tp, BSSGP_IE_IMSI);
 		osmo_mobile_identity_decode(&mi, mi_data, mi_len, false);
-		gbproxy_imsi_cache_update(nse, mi.imsi);
+		gbproxy_imsi_cache_update(nse, mi.imsi, CACHE_USAGE_PAGING);
 		break;
 	}
 	default:
@@ -685,26 +699,34 @@ static void bss_ptp_bvc_reset_notif(uint16_t nsei, uint16_t bvci, const struct g
 		bvc->cell->bss_bvc = bvc;
 	}
 
-	/* allocate (any missing) SGSN-side BVCs within the cell, and reset them */
+	/* Ensure we have the correct RA/CELL ID */
+	if (!gsm48_ra_equal(&bvc->cell->id.raid, ra_id)) {
+		LOGPBVC(bvc, LOGL_NOTICE, "RAID changed from %s to %s, updating cell\n", osmo_rai_name(&bvc->cell->id.raid), osmo_rai_name(ra_id));
+		memcpy(&bvc->cell->id.raid, ra_id, sizeof(*ra_id));
+	}
+	if (bvc->cell->id.cid != cell_id) {
+		LOGPBVC(bvc, LOGL_NOTICE, "CellID changed from %05d to %05d, updating cell\n", bvc->cell->id.cid, cell_id);
+		bvc->cell->id.cid = cell_id;
+	}
+
+	/* Reallocate SGSN-side BVCs of the cell, and reset them
+	 * Removing and reallocating is needed because the ra_id/cell_id might have changed */
 	hash_for_each(cfg->sgsn_nses, i, sgsn_nse, list) {
 		struct gbproxy_bvc *sgsn_bvc = gbproxy_bvc_by_bvci(sgsn_nse, bvci);
+		if (!sgsn_bvc)
+			sgsn_bvc = gbproxy_bvc_by_bvci_inactive(sgsn_nse, bvci);
 		if (sgsn_bvc)
-			OSMO_ASSERT(sgsn_bvc->cell == bvc->cell || !sgsn_bvc->cell);
+			gbproxy_bvc_free(sgsn_bvc);
 
-		if (!sgsn_bvc) {
-			sgsn_bvc = gbproxy_bvc_alloc(sgsn_nse, bvci);
-			OSMO_ASSERT(sgsn_bvc);
-
-			sgsn_bvc->cell = bvc->cell;
-			memcpy(&sgsn_bvc->raid, &bvc->cell->id.raid, sizeof(sgsn_bvc->raid));
-			sgsn_bvc->fi = bssgp_bvc_fsm_alloc_ptp_bss(sgsn_bvc, cfg->nsi, sgsn_nse->nsei,
-								   bvci, ra_id, cell_id);
-			OSMO_ASSERT(sgsn_bvc->fi);
-			bssgp_bvc_fsm_set_max_pdu_len(sgsn_bvc->fi, sgsn_nse->max_sdu_len);
-			bssgp_bvc_fsm_set_ops(sgsn_bvc->fi, &sgsn_ptp_bvc_fsm_ops, sgsn_bvc);
-
-			gbproxy_cell_add_sgsn_bvc(bvc->cell, sgsn_bvc);
-		}
+		sgsn_bvc = gbproxy_bvc_alloc(sgsn_nse, bvci);
+		OSMO_ASSERT(sgsn_bvc);
+		sgsn_bvc->cell = bvc->cell;
+		sgsn_bvc->fi = bssgp_bvc_fsm_alloc_ptp_bss(sgsn_bvc, cfg->nsi, sgsn_nse->nsei,
+							   bvci, ra_id, cell_id);
+		OSMO_ASSERT(sgsn_bvc->fi);
+		bssgp_bvc_fsm_set_max_pdu_len(sgsn_bvc->fi, sgsn_nse->max_sdu_len);
+		bssgp_bvc_fsm_set_ops(sgsn_bvc->fi, &sgsn_ptp_bvc_fsm_ops, sgsn_bvc);
+		gbproxy_cell_add_sgsn_bvc(bvc->cell, sgsn_bvc);
 	}
 
 	/* Trigger outbound BVC-RESET procedure toward each SGSN */
@@ -865,34 +887,6 @@ static int rx_bvc_reset_from_bss(struct gbproxy_nse *nse, struct msgb *msg, stru
 			bssgp_bvc_fsm_set_max_pdu_len(from_bvc->fi, nse->max_sdu_len);
 			bssgp_bvc_fsm_set_ops(from_bvc->fi, &bss_ptp_bvc_fsm_ops, from_bvc);
 		}
-#if 0
-		/* Could have moved to a different NSE */
-		if (!check_bvc_nsei(from_bvc, nsei)) {
-			LOGPBVC(from_bvc, LOGL_NOTICE, "moving bvc to NSE(%05u)\n", nsei);
-
-			struct gbproxy_nse *nse_new = gbproxy_nse_by_nsei(cfg, nsei, false);
-			if (!nse_new) {
-				LOGP(DGPRS, LOGL_NOTICE, "NSE(%05u) Got PtP BVC reset before signalling reset for "
-					"BVCI=%05u\n", bvci, nsei);
-				tx_status(nse, ns_bvci, BSSGP_CAUSE_PDU_INCOMP_STATE, NULL, msg);
-				return 0;
-			}
-
-			/* Move bvc to different NSE */
-			gbproxy_bvc_move(from_bvc, nse_new);
-		}
-#endif
-		/* FIXME: do we need this, if it happens within FSM? */
-		if (TLVP_PRES_LEN(tp, BSSGP_IE_CELL_ID, 8)) {
-			struct gprs_ra_id raid;
-			/* We have a Cell Identifier present in this
-			 * PDU, this means we can extend our local
-			 * state information about this particular cell
-			 * */
-			gsm48_parse_ra(&raid, TLVP_VAL(tp, BSSGP_IE_CELL_ID));
-			memcpy(&from_bvc->raid, &raid, sizeof(from_bvc->raid));
-			LOGPBVC(from_bvc, LOGL_INFO, "Cell ID %s\n", osmo_rai_name(&raid));
-		}
 	}
 	/* hand into FSM for further processing */
 	osmo_fsm_inst_dispatch(from_bvc->fi, BSSGP_BVCFSM_E_RX_RESET, msg);
@@ -979,6 +973,56 @@ static int gbprox_rx_rim_from_bss(struct tlv_parsed *tp, struct gbproxy_nse *nse
 	return gbprox_relay2nse(msg, sgsn->nse, 0);
 }
 
+/* Extract the TLLI from the PDU-in-error of the STATUS PDU (if available) */
+static int gbproxy_tlli_from_status_pdu(struct tlv_parsed *tp, uint32_t *tlli, char *log_pfx)
+{
+	int rc;
+	int pdu_len = TLVP_LEN(&tp[0], BSSGP_IE_PDU_IN_ERROR);
+	const uint8_t *pdu_data = TLVP_VAL(&tp[0], BSSGP_IE_PDU_IN_ERROR);
+	struct bssgp_normal_hdr *bgph = (struct bssgp_normal_hdr *)pdu_data;
+	struct tlv_parsed tp_inner;
+
+	/* TODO: Parse partial messages as well */
+	rc = gbproxy_decode_bssgp(bgph, pdu_len, &tp_inner, log_pfx);
+	if (rc < 0)
+		return rc;
+
+	if (TLVP_PRESENT(&tp_inner, BSSGP_IE_TLLI)) {
+		*tlli = osmo_load32be(TLVP_VAL(&tp_inner, BSSGP_IE_TLLI));
+	} else if (TLVP_PRESENT(&tp_inner, BSSGP_IE_TMSI)) {
+		/* we treat the TMSI like a TLLI and extract the NRI from it */
+		*tlli = osmo_load32be(TLVP_VAL(&tp_inner, BSSGP_IE_TMSI));
+		/* Convert the TMSI into a FOREIGN TLLI so it is routed appropriately */
+		*tlli = gprs_tmsi2tlli(*tlli, TLLI_FOREIGN);
+	} else {
+		return -ENOENT;
+	}
+
+	return 0;
+}
+
+/* Extract the BVCI from the PDU-in-error of the STATUS PDU (if available) */
+static int gbproxy_bvci_from_status_pdu(struct tlv_parsed *tp, uint16_t *bvci, char *log_pfx)
+{
+	int rc;
+	int pdu_len = TLVP_LEN(&tp[0], BSSGP_IE_PDU_IN_ERROR);
+	const uint8_t *pdu_data = TLVP_VAL(&tp[0], BSSGP_IE_PDU_IN_ERROR);
+	struct bssgp_normal_hdr *bgph = (struct bssgp_normal_hdr *)pdu_data;
+	struct tlv_parsed tp_inner;
+
+	/* TODO: Parse partial messages as well */
+	rc = gbproxy_decode_bssgp(bgph, pdu_len, &tp_inner, log_pfx);
+	if (rc < 0)
+		return rc;
+
+	if (TLVP_PRESENT(&tp_inner, BSSGP_IE_BVCI))
+		*bvci = ntohs(tlvp_val16_unal(&tp_inner, BSSGP_IE_BVCI));
+	else
+		return -ENOENT;
+
+	return 0;
+}
+
 /* Receive an incoming signalling message from a BSS-side NS-VC */
 static int gbprox_rx_sig_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uint16_t ns_bvci)
 {
@@ -1015,7 +1059,7 @@ static int gbprox_rx_sig_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 	rc = osmo_tlv_prot_parse(&osmo_pdef_bssgp, tp, ARRAY_SIZE(tp), pdu_type, bgph->data, data_len, 0, 0,
 				 DGPRS, log_pfx);
 	if (rc < 0) {
-		rate_ctr_inc(&nse->cfg->ctrg->ctr[GBPROX_GLOB_CTR_PROTO_ERR_BSS]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(nse->cfg->ctrg, GBPROX_GLOB_CTR_PROTO_ERR_BSS));
 		return tx_status_from_tlvp(nse, rc, msg);
 	}
 	/* hack to get both msg + tlv_parsed passed via osmo_fsm_inst_dispatch */
@@ -1069,9 +1113,36 @@ static int gbprox_rx_sig_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 		break;
 	}
 	case BSSGP_PDUT_STATUS:
-		/* FIXME: inspect the erroneous PDU IE (if any) and check
-		 * if we can extract a TLLI/RNI to route it to the correct SGSN */
+	{
+		struct gbproxy_sgsn *sgsn;
+		/* Check if the status needs to be terminated locally */
+		uint8_t cause = *TLVP_VAL(&tp[0], BSSGP_IE_CAUSE);
+
+		if (cause == BSSGP_CAUSE_UNKNOWN_BVCI || cause == BSSGP_CAUSE_BVCI_BLOCKED) {
+			/* Log and handle locally */
+			ptp_bvci = ntohs(tlvp_val16_unal(&tp[0], BSSGP_IE_BVCI));
+			LOGPNSE(nse, LOGL_ERROR, "Rx STATUS cause=0x%02x(%s) for PtP-BVC %05u\n", cause,
+				bssgp_cause_str(cause), ptp_bvci);
+			/* FIXME: Remove/block our BVC if present? */
+			break;
+		}
+
+		LOGPNSE(nse, LOGL_NOTICE, "Rx STATUS cause=0x%02x(%s) ", cause,
+			bssgp_cause_str(cause));
+
+		if (gbproxy_tlli_from_status_pdu(tp, &tlli, log_pfx) == 0)
+			sgsn = gbproxy_select_sgsn(nse->cfg, &tlli);
+		else
+			sgsn = gbproxy_select_sgsn(nse->cfg, NULL);
+
+		if (!sgsn) {
+			rc = -EINVAL;
+			break;
+		}
+
+		rc = gbprox_relay2nse(msg, sgsn->nse, 0);
 		break;
+	}
 	case BSSGP_PDUT_RAN_INFO:
 	case BSSGP_PDUT_RAN_INFO_REQ:
 	case BSSGP_PDUT_RAN_INFO_ACK:
@@ -1097,7 +1168,7 @@ static int gbprox_rx_sig_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 		const uint8_t *mi_data = TLVP_VAL(&tp[0], BSSGP_IE_IMSI);
 		uint8_t mi_len = TLVP_LEN(&tp[0], BSSGP_IE_IMSI);
 		osmo_mobile_identity_decode(&mi, mi_data, mi_len, false);
-		nse = gbproxy_nse_by_imsi(nse->cfg, mi.imsi);
+		nse = gbproxy_nse_by_imsi(nse->cfg, mi.imsi, CACHE_USAGE_PAGING);
 		if (!nse) {
 			return tx_status(nse, ns_bvci, BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
 		}
@@ -1105,6 +1176,26 @@ static int gbprox_rx_sig_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 		rc = gbprox_relay2nse(msg, nse, 0);
 		break;
 	}
+	case BSSGP_PDUT_MS_REGISTR_ENQ:
+	{
+		struct gbproxy_sgsn *sgsn;
+		struct osmo_mobile_identity mi;
+		const uint8_t *mi_data = TLVP_VAL(&tp[0], BSSGP_IE_IMSI);
+		uint8_t mi_len = TLVP_LEN(&tp[0], BSSGP_IE_IMSI);
+		osmo_mobile_identity_decode(&mi, mi_data, mi_len, false);
+
+		sgsn = gbproxy_select_sgsn(nse->cfg, NULL);
+		if (!sgsn) {
+			LOGP(DGPRS, LOGL_ERROR, "Could not find any SGSN, dropping message!\n");
+			rc = -EINVAL;
+			break;
+		}
+
+		gbproxy_imsi_cache_update(nse, mi.imsi, CACHE_USAGE_MS_REG_ENQ);
+
+		rc = gbprox_relay2nse(msg, sgsn->nse, 0);
+		break;
+	}
 	default:
 		LOGPNSE(nse, LOGL_ERROR, "Rx %s: Implementation missing\n", pdut_name);
 		break;
@@ -1113,7 +1204,7 @@ static int gbprox_rx_sig_from_bss(struct gbproxy_nse *nse, struct msgb *msg, uin
 	return rc;
 err_no_bvc:
 	LOGPNSE(nse, LOGL_ERROR, "Rx %s: cannot find BVC for BVCI=%05u\n", pdut_name, ptp_bvci);
-	rate_ctr_inc(&nse->cfg->ctrg->ctr[GBPROX_GLOB_CTR_INV_NSEI]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(nse->cfg->ctrg, GBPROX_GLOB_CTR_INV_NSEI));
 	return tx_status(nse, ns_bvci, BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
 }
 
@@ -1128,8 +1219,6 @@ static int gbprox_rx_paging(struct gbproxy_nse *sgsn_nse, struct msgb *msg, cons
 	int errctr = GBPROX_GLOB_CTR_PROTO_ERR_SGSN;
 	int i, j;
 
-	/* FIXME: Handle paging logic to only page each matching NSE */
-
 	if (TLVP_PRES_LEN(tp, BSSGP_IE_BVCI, 2)) {
 		uint16_t bvci = ntohs(tlvp_val16_unal(tp, BSSGP_IE_BVCI));
 		errctr = GBPROX_GLOB_CTR_OTHER_ERR;
@@ -1137,7 +1226,7 @@ static int gbprox_rx_paging(struct gbproxy_nse *sgsn_nse, struct msgb *msg, cons
 		if (!sgsn_bvc) {
 			LOGPNSE(sgsn_nse, LOGL_NOTICE, "Rx %s: unable to route: BVCI=%05u unknown\n",
 				pdut_name, bvci);
-			rate_ctr_inc(&cfg->ctrg->ctr[errctr]);
+			rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, errctr));
 			return -EINVAL;
 		}
 		LOGPBVC(sgsn_bvc, LOGL_INFO, "Rx %s: routing by BVCI\n", pdut_name);
@@ -1149,7 +1238,11 @@ static int gbprox_rx_paging(struct gbproxy_nse *sgsn_nse, struct msgb *msg, cons
 		/* iterate over all bvcs and dispatch the paging to each matching one */
 		hash_for_each(cfg->bss_nses, i, nse, list) {
 			hash_for_each(nse->bvcs, j, bss_bvc, list) {
-				if (gsm48_ra_equal(&bss_bvc->raid, &raid)) {
+				/* Skip BVCs without a cell (e.g. signalling) */
+				if (!bss_bvc->cell)
+					continue;
+
+				if (gsm48_ra_equal(&bss_bvc->cell->id.raid, &raid)) {
 					LOGPNSE(nse, LOGL_INFO, "Rx %s: routing to NSE (RAI match)\n",
 						pdut_name);
 					gbprox_relay2peer(msg, bss_bvc, ns_bvci);
@@ -1165,7 +1258,11 @@ static int gbprox_rx_paging(struct gbproxy_nse *sgsn_nse, struct msgb *msg, cons
 		/* iterate over all bvcs and dispatch the paging to each matching one */
 		hash_for_each(cfg->bss_nses, i, nse, list) {
 			hash_for_each(nse->bvcs, j, bss_bvc, list) {
-				gsm48_encode_ra(&lac, &bss_bvc->raid);
+				/* Skip BVCs without a cell (e.g. signalling) */
+				if (!bss_bvc->cell)
+					continue;
+
+				gsm48_encode_ra(&lac, &bss_bvc->cell->id.raid);
 				if (!memcmp(&lac, TLVP_VAL(tp, BSSGP_IE_LOCATION_AREA), 5)) {
 					LOGPNSE(nse, LOGL_INFO, "Rx %s: routing to NSE (LAI match)\n",
 						pdut_name);
@@ -1189,12 +1286,12 @@ static int gbprox_rx_paging(struct gbproxy_nse *sgsn_nse, struct msgb *msg, cons
 		}
 	} else {
 		LOGPNSE(sgsn_nse, LOGL_ERROR, "BSSGP PAGING: unable to route, missing IE\n");
-		rate_ctr_inc(&cfg->ctrg->ctr[errctr]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, errctr));
 	}
 
 	if (n_nses == 0) {
 		LOGPNSE(sgsn_nse, LOGL_ERROR, "BSSGP PAGING: unable to route, no destination found\n");
-		rate_ctr_inc(&cfg->ctrg->ctr[errctr]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, errctr));
 		return -EINVAL;
 	}
 	return 0;
@@ -1217,7 +1314,7 @@ static int rx_bvc_reset_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, str
 		from_bvc = gbproxy_bvc_by_bvci(nse, ptp_bvci);
 		if (!from_bvc) {
 			LOGPNSE(nse, LOGL_ERROR, "Rx BVC-RESET BVCI=%05u: Cannot find BVC\n", ptp_bvci);
-			rate_ctr_inc(&nse->cfg->ctrg->ctr[GBPROX_GLOB_CTR_INV_BVCI]);
+			rate_ctr_inc(rate_ctr_group_get_ctr(nse->cfg->ctrg, GBPROX_GLOB_CTR_INV_BVCI));
 			return tx_status(nse, ns_bvci, BSSGP_CAUSE_UNKNOWN_BVCI, &ptp_bvci, msg);
 		}
 		osmo_fsm_inst_dispatch(from_bvc->fi, BSSGP_BVCFSM_E_RX_RESET, msg);
@@ -1296,7 +1393,6 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 	uint16_t bvci;
 	char log_pfx[32];
 	int rc = 0;
-	int cause;
 	int i;
 	bool paging_bc = false;
 
@@ -1325,7 +1421,7 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 				 DGPRS, log_pfx);
 	if (rc < 0) {
 		rc = tx_status_from_tlvp(nse, rc, msg);
-		rate_ctr_inc(&cfg->ctrg->ctr[GBPROX_GLOB_CTR_PROTO_ERR_SGSN]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, GBPROX_GLOB_CTR_PROTO_ERR_SGSN));
 		return rc;
 	}
 	/* hack to get both msg + tlv_parsed passed via osmo_fsm_inst_dispatch */
@@ -1347,8 +1443,12 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 	case BSSGP_PDUT_BVC_BLOCK_ACK:
 		bvci = ntohs(tlvp_val16_unal(&tp[0], BSSGP_IE_BVCI));
 		sgsn_bvc = gbproxy_bvc_by_bvci(nse, bvci);
-		if (!sgsn_bvc)
-			goto err_no_bvc;
+		if (!sgsn_bvc) {
+			/* Check if BVC was blocked before */
+			sgsn_bvc = gbproxy_bvc_by_bvci_inactive(nse, bvci);
+			if (!sgsn_bvc)
+				goto err_no_bvc;
+		}
 		rc = osmo_fsm_inst_dispatch(sgsn_bvc->fi, BSSGP_BVCFSM_E_RX_BLOCK_ACK, msg);
 		break;
 	case BSSGP_PDUT_BVC_UNBLOCK_ACK:
@@ -1379,7 +1479,7 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 		const uint8_t *mi_data = TLVP_VAL(&tp[0], BSSGP_IE_IMSI);
 		uint8_t mi_len = TLVP_LEN(&tp[0], BSSGP_IE_IMSI);
 		osmo_mobile_identity_decode(&mi, mi_data, mi_len, false);
-		gbproxy_imsi_cache_update(nse, mi.imsi);
+		gbproxy_imsi_cache_update(nse, mi.imsi, CACHE_USAGE_PAGING);
 		/* fall through */
 	}
 	case BSSGP_PDUT_PAGING_CS:
@@ -1387,20 +1487,55 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 		rc = gbprox_rx_paging(nse, msg, pdut_name, &tp[0], ns_bvci, paging_bc);
 		break;
 	case BSSGP_PDUT_STATUS:
-		/* Some exception has occurred */
-		cause = *TLVP_VAL(&tp[0], BSSGP_IE_CAUSE);
-		LOGPNSE(nse, LOGL_NOTICE, "Rx STATUS cause=0x%02x(%s) ", cause,
+	{
+		struct gbproxy_nse *nse_peer;
+		uint32_t tlli;
+
+		/* Check if the status needs to be terminated locally */
+		uint8_t cause = *TLVP_VAL(&tp[0], BSSGP_IE_CAUSE);
+
+		if (cause == BSSGP_CAUSE_UNKNOWN_BVCI || cause == BSSGP_CAUSE_BVCI_BLOCKED) {
+			/* Log and handle locally, BVCI should be present for these causes */
+			if (!TLVP_PRESENT(&tp[0], BSSGP_IE_BVCI)) {
+				LOGPNSE(nse, LOGL_ERROR, "Rx STATUS cause=0x%02x(%s), but BVCI is missing\n", cause,
+					bssgp_cause_str(cause));
+				break;
+			}
+			uint16_t ptp_bvci = ntohs(tlvp_val16_unal(&tp[0], BSSGP_IE_BVCI));
+			LOGPNSE(nse, LOGL_ERROR, "Rx STATUS cause=0x%02x(%s) for PtP-BVC %05u\n", cause,
+				bssgp_cause_str(cause), ptp_bvci);
+			/* FIXME: Remove/block the other BSS/SGSN BVCs if present? */
+			break;
+		}
+
+		LOGPNSE(nse, LOGL_NOTICE, "Rx STATUS cause=0x%02x(%s)\n", cause,
 			bssgp_cause_str(cause));
-		if (TLVP_PRES_LEN(&tp[0], BSSGP_IE_BVCI, 2)) {
-			bvci = ntohs(tlvp_val16_unal(&tp[0], BSSGP_IE_BVCI));
-			LOGPC(DGPRS, LOGL_NOTICE, "BVCI=%05u\n", bvci);
-			sgsn_bvc = gbproxy_bvc_by_bvci(nse, bvci);
-			/* don't send STATUS in response to STATUS if !bvc */
-			if (sgsn_bvc && sgsn_bvc->cell && sgsn_bvc->cell->bss_bvc)
-				rc = gbprox_relay2peer(msg, sgsn_bvc->cell->bss_bvc, ns_bvci);
-		} else
-			LOGPC(DGPRS, LOGL_NOTICE, "\n");
+
+
+		if (gbproxy_bvci_from_status_pdu(tp, &bvci, log_pfx) == 0 && bvci != 0) {
+			struct gbproxy_cell *cell = gbproxy_cell_by_bvci(cfg, bvci);
+
+			if ((!cell || !cell->bss_bvc || !cell->bss_bvc->nse)) {
+				LOGPNSE(nse, LOGL_ERROR, "Rx STATUS cause=0x%02x(%s), but can't find NSE for cell\n",
+					cause, bssgp_cause_str(cause));
+				break;
+			}
+
+			return gbprox_relay2nse(msg, cell->bss_bvc->nse, 0);
+		}
+
+		/* We can only forward this TLLI if it's in the cache (which only happens on suspend/resume) */
+		if (gbproxy_tlli_from_status_pdu(tp, &tlli, log_pfx) == 0) {
+			nse_peer = gbproxy_nse_by_tlli(cfg, tlli);
+			if (nse_peer)
+				return gbprox_relay2nse(msg, nse_peer, 0);
+		}
+
+		LOGPNSE(nse, LOGL_ERROR, "Unable to handle STATUS cause=0x%02x(%s)\n", cause,
+			bssgp_cause_str(cause));
+
 		break;
+	}
 	/* those only exist in the SGSN -> BSS direction */
 	case BSSGP_PDUT_SUSPEND_ACK:
 	case BSSGP_PDUT_SUSPEND_NACK:
@@ -1436,9 +1571,29 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 	case BSSGP_PDUT_RAN_INFO_ERROR:
 	case BSSGP_PDUT_RAN_INFO_APP_ERROR:
 		rc = gbprox_rx_rim_from_sgsn(tp, nse, msg, log_pfx, pdut_name);
+		break;
+	case BSSGP_PDUT_MS_REGISTR_ENQ_RESP:
+	{
+		struct gbproxy_nse *nse_peer;
+		struct osmo_mobile_identity mi;
+		const uint8_t *mi_data = TLVP_VAL(&tp[0], BSSGP_IE_IMSI);
+		uint8_t mi_len = TLVP_LEN(&tp[0], BSSGP_IE_IMSI);
+		osmo_mobile_identity_decode(&mi, mi_data, mi_len, false);
+		nse_peer = gbproxy_nse_by_imsi(cfg, mi.imsi, CACHE_USAGE_MS_REG_ENQ);
+		if (!nse_peer) {
+			LOGPNSE(nse, LOGL_ERROR, "Rx %s: Cannot find NSE\n", pdut_name);
+			return tx_status(nse, ns_bvci, BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
+		} else if (nse_peer->sgsn_facing) {
+			LOGPNSE(nse, LOGL_ERROR, "Forwarding %s failed: IMSI cache contains SGSN NSE", pdut_name);
+			return tx_status(nse, ns_bvci, BSSGP_CAUSE_PROTO_ERR_UNSPEC, NULL, msg);
+		}
+		gbproxy_imsi_cache_remove(cfg, mi.imsi, CACHE_USAGE_MS_REG_ENQ);
+		gbprox_relay2nse(msg, nse_peer, ns_bvci);
+		break;
+	}
 	default:
 		LOGPNSE(nse, LOGL_NOTICE, "Rx %s: Not supported\n", pdut_name);
-		rate_ctr_inc(&cfg->ctrg->ctr[GBPROX_GLOB_CTR_PROTO_ERR_SGSN]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, GBPROX_GLOB_CTR_PROTO_ERR_SGSN));
 		rc = tx_status(nse, ns_bvci, BSSGP_CAUSE_PROTO_ERR_UNSPEC, NULL, msg);
 		break;
 	}
@@ -1446,8 +1601,8 @@ static int gbprox_rx_sig_from_sgsn(struct gbproxy_nse *nse, struct msgb *msg, ui
 	return rc;
 
 err_no_bvc:
-	LOGPNSE(nse, LOGL_ERROR, "Rx %s: Cannot find BVC\n", pdut_name);
-	rate_ctr_inc(&cfg->ctrg-> ctr[GBPROX_GLOB_CTR_INV_RAI]);
+	LOGPNSE(nse, LOGL_ERROR, "Rx %s: Cannot find BVC %05u\n", pdut_name, bvci);
+	rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, GBPROX_GLOB_CTR_INV_BVCI));
 	return tx_status(nse, ns_bvci, BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
 }
 
@@ -1514,8 +1669,6 @@ int gbprox_rcvmsg(void *ctx, struct msgb *msg)
 /*  TODO: What about handling:
  * 	GPRS_NS2_AFF_CAUSE_VC_FAILURE,
 	GPRS_NS2_AFF_CAUSE_VC_RECOVERY,
-	GPRS_NS2_AFF_CAUSE_FAILURE,
-	GPRS_NS2_AFF_CAUSE_RECOVERY,
 	osmocom own causes
 	GPRS_NS2_AFF_CAUSE_SNS_CONFIGURED,
 	GPRS_NS2_AFF_CAUSE_SNS_FAILURE,
@@ -1523,13 +1676,9 @@ int gbprox_rcvmsg(void *ctx, struct msgb *msg)
 
 void gprs_ns_prim_status_cb(struct gbproxy_config *cfg, struct osmo_gprs_ns2_prim *nsp)
 {
-	/* TODO: bss nsei available/unavailable  bssgp_tx_simple_bvci(BSSGP_PDUT_BVC_BLOCK, nsvc->nsei, bvc->bvci, 0);
-	 * TODO: sgsn nsei available/unavailable
-	 * TODO: Update MTU
-	 */
-
+	int i;
 	struct gbproxy_bvc *bvc;
-	struct gbproxy_nse *sgsn_nse;
+	struct gbproxy_nse *nse;
 
 	switch (nsp->u.status.cause) {
 	case GPRS_NS2_AFF_CAUSE_SNS_FAILURE:
@@ -1538,39 +1687,61 @@ void gprs_ns_prim_status_cb(struct gbproxy_config *cfg, struct osmo_gprs_ns2_pri
 
 	case GPRS_NS2_AFF_CAUSE_RECOVERY:
 		LOGP(DGPRS, LOGL_NOTICE, "NS-NSE %d became available\n", nsp->nsei);
-		sgsn_nse = gbproxy_nse_by_nsei(cfg, nsp->nsei, NSE_F_SGSN);
-		if (sgsn_nse) {
+		nse = gbproxy_nse_by_nsei(cfg, nsp->nsei, NSE_F_SGSN);
+		if (nse) {
+			nse->alive = true;
+			// Update the NSE max SDU len
+			nse->max_sdu_len = nsp->u.status.mtu;
+
 			uint8_t cause = BSSGP_CAUSE_OML_INTERV;
-			bvc = gbproxy_bvc_by_bvci(sgsn_nse, 0);
-			if (bvc)
+			bvc = gbproxy_bvc_by_bvci(nse, 0);
+			if (bvc) {
+				bssgp_bvc_fsm_set_max_pdu_len(bvc->fi, nse->max_sdu_len);
 				osmo_fsm_inst_dispatch(bvc->fi, BSSGP_BVCFSM_E_REQ_RESET, &cause);
+			}
 		}
 		break;
 	case GPRS_NS2_AFF_CAUSE_FAILURE:
-#if 0
-		if (gbproxy_is_sgsn_nsei(cfg, nsp->nsei)) {
-			/* sgsn */
-			/* TODO: BSVC: block all PtP towards bss */
-			rate_ctr_inc(&cfg->ctrg->
-				     ctr[GBPROX_GLOB_CTR_RESTART_RESET_SGSN]);
+		nse = gbproxy_nse_by_nsei(cfg, nsp->nsei, NSE_F_BSS | NSE_F_SGSN);
+		if (!nse) {
+			LOGP(DGPRS, LOGL_ERROR, "Unknown NSE(%05d) became unavailable\n", nsp->nsei);
+			break;
+		}
+
+		nse->alive = false;
+		if (nse->sgsn_facing) {
+			struct hlist_node *ntmp;
+			/* SGSN */
+			/* TODO: When to block all PtP towards bss? Only if all SGSN are down? */
+			hash_for_each_safe(nse->bvcs, i, ntmp, bvc, list) {
+				if (bvc->bvci == 0)
+					continue;
+				gbproxy_bvc_free(bvc);
+			}
+			rate_ctr_inc(rate_ctr_group_get_ctr(cfg->ctrg, GBPROX_GLOB_CTR_RESTART_RESET_SGSN));
 		} else {
-			/* bss became unavailable
-			 * TODO: Block all BVC belonging to that NSE */
-			bvc = gbproxy_bvc_by_nsei(cfg, nsp->nsei);
-			if (!bvc) {
-				/* TODO: use primitive name + status cause name */
-				LOGP(DGPRS, LOGL_NOTICE, "Received ns2 primitive %d for unknown bvc NSEI=%u\n",
-				     nsp->u.status.cause, nsp->nsei);
-				break;
+			/* BSS became unavailable
+			 * Block matching PtP-BVCs on SGSN-side */
+			hash_for_each(nse->bvcs, i, bvc, list) {
+				if (bvc->bvci == 0)
+					continue;
+				/* Get BVC for each SGSN and send block request */
+				struct gbproxy_cell *cell = bvc->cell;
+				for (int j = 0; j < GBPROXY_MAX_NR_SGSN; j++) {
+					struct gbproxy_bvc *sgsn_bvc = cell->sgsn_bvc[j];
+					if (!sgsn_bvc)
+						continue;
+
+					/* Block BVC, indicate BSS equipment failure */
+					uint8_t cause = BSSGP_CAUSE_EQUIP_FAIL;
+					osmo_fsm_inst_dispatch(sgsn_bvc->fi, BSSGP_BVCFSM_E_REQ_BLOCK, &cause);
+					sgsn_bvc->inactive = true;
+				}
 			}
 
-			if (!bvc->blocked)
-				break;
-			hash_for_each(cfg->sgsn_nses, _sgsn, sgsn_nse, list) {
-				bssgp_tx_simple_bvci(BSSGP_PDUT_BVC_BLOCK, sgsn_nse->nsei, bvc->bvci, 0);
-			}
+			/* This frees the BVCs for us as well */
+			gbproxy_nse_free(nse);
 		}
-#endif
 		LOGP(DGPRS, LOGL_NOTICE, "NS-NSE %d became unavailable\n", nsp->nsei);
 		break;
 	default:
