@@ -20,7 +20,7 @@
  *
  */
 
-#include <osmocom/sgsn/gb_proxy.h>
+#include <osmocom/gbproxy/gb_proxy.h>
 
 #include "debug.h"
 
@@ -62,7 +62,18 @@ struct gbproxy_bvc *gbproxy_bvc_by_bvci(struct gbproxy_nse *nse, uint16_t bvci)
 {
 	struct gbproxy_bvc *bvc;
 	hash_for_each_possible(nse->bvcs, bvc, list, bvci) {
-		if (bvc->bvci == bvci)
+		if (bvc->bvci == bvci && bvc->inactive == false)
+			return bvc;
+	}
+	return NULL;
+}
+
+/* Find the gbproxy_bvc by its BVCI. There can only be one match */
+struct gbproxy_bvc *gbproxy_bvc_by_bvci_inactive(struct gbproxy_nse *nse, uint16_t bvci)
+{
+	struct gbproxy_bvc *bvc;
+	hash_for_each_possible(nse->bvcs, bvc, list, bvci) {
+		if (bvc->bvci == bvci && bvc->inactive == true)
 			return bvc;
 	}
 	return NULL;
@@ -70,6 +81,7 @@ struct gbproxy_bvc *gbproxy_bvc_by_bvci(struct gbproxy_nse *nse, uint16_t bvci)
 
 struct gbproxy_bvc *gbproxy_bvc_alloc(struct gbproxy_nse *nse, uint16_t bvci)
 {
+	char idbuf[64];
 	struct gbproxy_bvc *bvc;
 	OSMO_ASSERT(nse);
 	struct gbproxy_config *cfg = nse->cfg;
@@ -79,12 +91,17 @@ struct gbproxy_bvc *gbproxy_bvc_alloc(struct gbproxy_nse *nse, uint16_t bvci)
 	if (!bvc)
 		return NULL;
 
+	snprintf(idbuf, sizeof(idbuf), "BVC%05u-NSE%05u", bvci,
+		 nse->nsei);
+	osmo_identifier_sanitize_buf(idbuf, NULL, '_');
 	bvc->bvci = bvci;
+	bvc->inactive = false;
 	bvc->ctrg = rate_ctr_group_alloc(bvc, &bvc_ctrg_desc, (nse->nsei << 16) | bvci);
 	if (!bvc->ctrg) {
 		talloc_free(bvc);
 		return NULL;
 	}
+	rate_ctr_group_set_name(bvc->ctrg, idbuf);
 	bvc->nse = nse;
 
 	hash_add(nse->bvcs, &bvc->list, bvc->bvci);
@@ -98,8 +115,6 @@ struct gbproxy_bvc *gbproxy_bvc_alloc(struct gbproxy_nse *nse, uint16_t bvci)
 
 void gbproxy_bvc_free(struct gbproxy_bvc *bvc)
 {
-	struct gbproxy_cell *cell;
-
 	if (!bvc)
 		return;
 
@@ -112,21 +127,10 @@ void gbproxy_bvc_free(struct gbproxy_bvc *bvc)
 
 	osmo_fsm_inst_free(bvc->fi);
 
-	cell = bvc->cell;
-	if (cell) {
-		int i;
-
-		if (cell->bss_bvc == bvc)
-			cell->bss_bvc = NULL;
-
-		/* we could also be a SGSN-side BVC */
-		for (i = 0; i < ARRAY_SIZE(cell->sgsn_bvc); i++) {
-			if (cell->sgsn_bvc[i] == bvc)
-				cell->sgsn_bvc[i] = NULL;
-		}
+	if (bvc->cell) {
+		gbproxy_cell_cleanup_bvc(bvc->cell, bvc);
 		bvc->cell = NULL;
 	}
-
 	talloc_free(bvc);
 }
 
@@ -209,6 +213,21 @@ struct gbproxy_cell *gbproxy_cell_by_cellid(struct gbproxy_config *cfg, const st
 	return NULL;
 }
 
+void gbproxy_cell_cleanup_bvc(struct gbproxy_cell *cell, struct gbproxy_bvc *bvc)
+{
+	int i;
+
+	if (cell->bss_bvc == bvc)
+		return gbproxy_cell_free(cell);
+
+	/* we could also be a SGSN-side BVC */
+	for (i = 0; i < ARRAY_SIZE(cell->sgsn_bvc); i++) {
+		if (cell->sgsn_bvc[i] == bvc)
+			cell->sgsn_bvc[i] = NULL;
+	}
+
+}
+
 void gbproxy_cell_free(struct gbproxy_cell *cell)
 {
 	unsigned int i;
@@ -347,23 +366,23 @@ static inline uint16_t _checksum_imsi(const char *imsi)
 	return osmo_crc16(0, (const uint8_t *)imsi, len);
 }
 
-static inline struct gbproxy_imsi_cache_entry *_get_imsi_entry(struct gbproxy_config *cfg, const char *imsi)
+static inline struct gbproxy_imsi_cache_entry *_get_imsi_entry(struct gbproxy_config *cfg, const char *imsi, enum cache_usage_type usage)
 {
 	struct gbproxy_imsi_cache_entry *cache_entry;
 	uint16_t imsi_hash = _checksum_imsi(imsi);
 
 	hash_for_each_possible(cfg->imsi_cache.entries, cache_entry, list, imsi_hash) {
-		if (!strncmp(cache_entry->imsi, imsi, sizeof(cache_entry->imsi)))
+		if (!strncmp(cache_entry->imsi, imsi, sizeof(cache_entry->imsi)) && cache_entry->usage == usage)
 			return cache_entry;
 	}
 	return NULL;
 }
 
-void gbproxy_imsi_cache_update(struct gbproxy_nse *nse, const char *imsi)
+void gbproxy_imsi_cache_update(struct gbproxy_nse *nse, const char *imsi, enum cache_usage_type usage)
 {
 	struct gbproxy_config *cfg = nse->cfg;
 	struct timespec now;
-	struct gbproxy_imsi_cache_entry *cache_entry = _get_imsi_entry(cfg, imsi);
+	struct gbproxy_imsi_cache_entry *cache_entry = _get_imsi_entry(cfg, imsi, usage);
 	uint16_t imsi_hash = _checksum_imsi(imsi);
 
 	osmo_clock_gettime(CLOCK_MONOTONIC, &now);
@@ -378,6 +397,7 @@ void gbproxy_imsi_cache_update(struct gbproxy_nse *nse, const char *imsi)
 	cache_entry = talloc_zero(cfg, struct gbproxy_imsi_cache_entry);
 	OSMO_STRLCPY_ARRAY(cache_entry->imsi, imsi);
 	cache_entry->nse = nse;
+	cache_entry->usage = usage;
 	cache_entry->tstamp = now.tv_sec;
 	hash_add(cfg->imsi_cache.entries, &cache_entry->list, imsi_hash);
 }
@@ -396,14 +416,14 @@ static void _imsi_cache_remove_nse(struct gbproxy_nse *nse) {
 	}
 }
 
-void gbproxy_imsi_cache_remove(struct gbproxy_config *cfg, const char *imsi)
+void gbproxy_imsi_cache_remove(struct gbproxy_config *cfg, const char *imsi, enum cache_usage_type usage)
 {
 	struct gbproxy_imsi_cache_entry *imsi_cache;
 	struct hlist_node *tmp;
 	uint16_t imsi_hash = _checksum_imsi(imsi);
 
 	hash_for_each_possible_safe(cfg->imsi_cache.entries, imsi_cache, tmp, list, imsi_hash) {
-		if (!(strncmp(imsi_cache->imsi, imsi, sizeof(imsi_cache->imsi)))) {
+		if (!(strncmp(imsi_cache->imsi, imsi, sizeof(imsi_cache->imsi))) && imsi_cache->usage == usage) {
 			hash_del(&imsi_cache->list);
 			talloc_free(imsi_cache);
 			return;
@@ -546,13 +566,13 @@ struct gbproxy_nse *gbproxy_nse_by_tlli(struct gbproxy_config *cfg, uint32_t tll
 	return NULL;
 }
 
-struct gbproxy_nse *gbproxy_nse_by_imsi(struct gbproxy_config *cfg, const char *imsi)
+struct gbproxy_nse *gbproxy_nse_by_imsi(struct gbproxy_config *cfg, const char *imsi, enum cache_usage_type usage)
 {
 	struct gbproxy_imsi_cache_entry *imsi_cache;
 	uint16_t imsi_hash = _checksum_imsi(imsi);
 
 	hash_for_each_possible(cfg->imsi_cache.entries, imsi_cache, list, imsi_hash) {
-		if (!strncmp(imsi_cache->imsi, imsi, sizeof(imsi_cache->imsi)))
+		if (!strncmp(imsi_cache->imsi, imsi, sizeof(imsi_cache->imsi)) && imsi_cache->usage == usage)
 			return imsi_cache->nse;
 	}
 	return NULL;
@@ -625,7 +645,7 @@ void gbproxy_sgsn_free(struct gbproxy_sgsn *sgsn)
 	if (!sgsn)
 		return;
 
-	OSMO_ASSERT(sgsn->nse)
+	OSMO_ASSERT(sgsn->nse);
 
 	_nse_free(sgsn->nse);
 	_sgsn_free(sgsn);
@@ -696,6 +716,9 @@ struct gbproxy_sgsn *gbproxy_sgsn_by_nri(struct gbproxy_config *cfg, uint16_t nr
 	OSMO_ASSERT(cfg);
 
 	llist_for_each_entry(sgsn, &cfg->sgsns, list) {
+		if (!sgsn->nse->alive)
+			continue;
+
 		if (osmo_nri_v_matches_ranges(nri, sgsn->pool.nri_ranges)) {
 			/* Also check if the NRI we're looking for is a NULL NRI */
 			if (null_nri) {
@@ -711,7 +734,7 @@ struct gbproxy_sgsn *gbproxy_sgsn_by_nri(struct gbproxy_config *cfg, uint16_t nr
 	return NULL;
 }
 
-/*! Select a pseudo-random SGSN for a given TLLI, ignoring any SGSN that is not accepting connections
+/*! Select a pseudo-random SGSN for a given TLLI, ignoring any SGSN that is not accepting connections or down
  *  \param[in] cfg The gbproxy configuration
  *  \param[in] sgsn_avoid If not NULL then avoid this SGSN when selecting a new one. Use for load redistribution
  *  \param[in] tlli The tlli to choose an SGSN for. The same tlli will map to the same SGSN as long as no SGSN is
@@ -743,7 +766,7 @@ struct gbproxy_sgsn *gbproxy_sgsn_by_tlli(struct gbproxy_config *cfg, struct gbp
 
 	/* Get the first enabled SGSN after index */
 	llist_for_each_entry(sgsn, &cfg->sgsns, list) {
-		if (i >= index && sgsn->pool.allow_attach) {
+		if (i >= index && sgsn->pool.allow_attach && sgsn->nse->alive) {
 			return sgsn;
 		}
 		i++;
@@ -753,7 +776,7 @@ struct gbproxy_sgsn *gbproxy_sgsn_by_tlli(struct gbproxy_config *cfg, struct gbp
 	llist_for_each_entry(sgsn, &cfg->sgsns, list) {
 		if (i >= index) {
 			break;
-		} else if (sgsn->pool.allow_attach) {
+		} else if (sgsn->pool.allow_attach && sgsn->nse->alive) {
 			return sgsn;
 		}
 		i++;
@@ -761,3 +784,19 @@ struct gbproxy_sgsn *gbproxy_sgsn_by_tlli(struct gbproxy_config *cfg, struct gbp
 
 	return NULL;
 }
+
+/*! Return the first available gbproxy_sgsn
+ *  \param[in] cfg proxy in which we operate
+ *  \return The SGSN, NULL if no matching SGSN could be found
+ */
+struct gbproxy_sgsn *gbproxy_sgsn_by_available(struct gbproxy_config *cfg)
+{
+	struct gbproxy_sgsn *sgsn;
+	OSMO_ASSERT(cfg);
+
+	llist_for_each_entry(sgsn, &cfg->sgsns, list)
+		if (sgsn->nse->alive &&sgsn->pool.allow_attach)
+			return sgsn;
+
+	return NULL;
+}
\ No newline at end of file
